## Display the inferred PCFGs

source("misc-lib.R")
source("sort.R")
source("HSFL-Shared.R")
library(boa) # for HPD intervals; Chen, M-H. and Shao, Q-M. (1999). Monte Carlo estimation of Bayesian credible and HPD intervals. Journal of Computational and Graphical Statistics, 8(1), 69-92. 
library(Hmisc)

hpd.alpha <- 0.05 # 95% HPD intervals

language <- "LAM-QEF-EOQ"
#  language <- "LAM-QEF"

model <- paste("../Model-Vicare/run/CONCEPT_ALL__LANGUAGE_Language-", language, "__INHYT_RAW__MODEL_parameterized-RR-noBeta", sep="")

## TODO: We may have to replace "" with " in these files
g <- read.table(paste(model, "/grammar-samples.txt", sep=""), header=T)

# get rid of burn-in
g <- g[-1:-500,]


D <- NULL
for(col in 3:ncol(g)) {

	gsorted <- g[order(g[1,], decreasing=T),col]

	g[,col] <- as.numeric(as.character(g[,col]))
	h <- boa.hpd(g[,col], hpd.alpha)
	mean <- mean(g[,col],na.rm=T)
	map  <- as.numeric(as.character(gsorted[1]))
	
	D <- rbind(D, data.frame(name=names(g)[col], mean=mean, MAP=map, hpd.lower=h[1], hpd.upper=h[2]))
}


# Fix the goddamn names
D <- D[D$mean > 0,] # these are zero for things we don't want
D$name <- as.character(D$name)
Nmap <- NULL
Nmap[D$name] <- D$name
Nmap["BOOL.to..exists..OBJ..BOOL.SET."] <- "BOOL to (exists F SET)"
Nmap["BOOL.to..forall..OBJ..BOOL.SET."] <- "BOOL to (forall F SET)"
Nmap["BOOL.to..only.satisfier..SET.OBJ.OBJ..BOOL.SET."] <- "BOOL to (only-satisfier OBJECT F SET)"
Nmap["BOOL.to..iff..BOOL.BOOL."] <- "BOOL to (iff BOOL BOOL)"
Nmap["BOOL.to..and..BOOL.BOOL."] <- "BOOL to (and BOOL BOOL)"
Nmap["BOOL.to..or..BOOL.BOOL."] <- "BOOL to (or BOOL BOOL)"
Nmap["BOOL.to..not..BOOL."] <- "BOOL to (not BOOL)"
Nmap["BOOL.to..OBJ..BOOL.SET.OBJ."] <- "BOOL to (F OBJECT)"
Nmap["SET.OBJ.to.BV"] <- "OBJECT to BV"
Nmap["SET.OBJ.to.x"] <- "OBJECT to x"
Nmap["SIZE.to.size1.."] <- "SIZE to size1?"
Nmap["SIZE.to.size2.."] <- "SIZE to size2?"
Nmap["SIZE.to.size3.."] <- "SIZE to size3?"
Nmap["OBJ..BOOL.to...lambda..SET.OBJ.BOOL.1."] <- "F to (lambda x_i BOOL)"
Nmap["OBJ..BOOL.to.SIZE"] <- "F to SIZE"
Nmap["OBJ..BOOL.to.SHAPE"] <- "F to SHAPE"
Nmap["OBJ..BOOL.to.COLOR"] <- "F to COLOR"
Nmap["SHAPE.to.triangle."] <- "SHAPE to triangle?"
Nmap["SHAPE.to.triangle.."] <- "SHAPE to triangle?"
Nmap["SHAPE.to.rectangle.."] <- "SHAPE to rectangle?"
Nmap["SHAPE.to.circle.."] <- "SHAPE to circle?"
Nmap["COLOR.to.yellow.."] <- "COLOR to yellow?"
Nmap["COLOR.to.green.."] <- "COLOR to green?"
Nmap["COLOR.to.blue.."] <- "COLOR to blue?"
Nmap["FUNCTION.to..lambda..S...lambda..x...falsethunk..."] <- "START to (lambda S x false)"
Nmap["FUNCTION.to..lambda..S...lambda..x...truethunk..."] <- "START to (lambda S x true)"
Nmap["FUNCTION.to..lambda..S...lambda..x..BOOL.."] <- "START to (lambda S x BOOL)"
Nmap["FUNCTION.to..lambda..S...lambda..x..ORCLAUSE.."] <- "START to (lambda S x ORCLAUSE)"
Nmap["FUNCTION.to..lambda..S...lambda..x..ANDCLAUSE.."] <- "START to (lambda S x ANDCLAUSE)"

Nmap["SET.to..non.xes.x.S."] <- "SET to (non-Xes S)"
Nmap["SET.to.S "] <- "SET to S"
Nmap["OBJxOBJ..BOOL.to.eqv.SFL.object."] <- "RELN to eqv-object?"
Nmap["OBJxOBJ..BOOL.to.size.."] <- "RELN to size>?"
Nmap["OBJxOBJ..BOOL.to.size..."] <- "RELN to size>=?"
Nmap["OBJxOBJ..BOOL.to.eqv.size."] <- "RELN to eqv-size?"
Nmap["OBJxOBJ..BOOL.to.eqv.shape."] <- "RELN to eqv-shape?"
Nmap["OBJxOBJ..BOOL.to.eqv.color."] <- "RELN to eqv-color?"
Nmap["BOOL.to..implies..BOOL.BOOL."] <- "BOOL to (implies BOOL BOOL)"

Nmap["PRIM.to..OBJ..BOOL.x."] <- "PRIM to (F x)"
Nmap["PRIM.to..not...OBJ..BOOL.x.."] <- "PRIM to (not (F x))"
Nmap["ANDCLAUSE.to..and..PRIM.ANDCLAUSE."] <- "ANDCLAUSE to (and PRIM AND-CLAUSE)"
Nmap["ANDCLAUSE.to.PRIM"] <- "ANDCLAUSE to PRIM"
Nmap["ORCLAUSE.to..or..ANDCLAUSE.ORCLAUSE."] <- "ORCLAUSE to (or ANDCLAUSE ORCLAUSE)"

Nmap["ANDCLAUSE.to..and..ORCLAUSE.ANDCLAUSE."] <- "ANDCLAUSE to (and ORCLAUSE AND-CLAUSE)"
Nmap["ORCLAUSE.to..or..PRIM.ORCLAUSE."] <- "ORCLAUSE to (or PRIM ORCLAUSE)"
Nmap["ORCLAUSE.to.PRIM"] <- "ORCLAUSE to PRIM"
Nmap["ANDCLAUSE.to.ORCLAUSE"] <- "ANDCLAUSE to ORCLAUSE"

Nmap["BOOL.to..exists.one.or.fewer..OBJ..BOOL.SET."] <- "BOOL to (exists-one-or-fewer F SET)"
Nmap["BOOL.to..satisfiers...OBJ..BOOL.OBJ..BOOL.SET."] <- "BOOL to (satisfiers> F F SET)"
Nmap["PRED.LAMBDA.to...lambda..OBJ..BOOL.BOOL.10."] <- "PRED-LAMBDA to (lambda P . BOOL)"
Nmap["BOOL.to..satisfiers...OBJ..BOOL.OBJ..BOOL.SET..1"] <- "BOOL to (satisfiers= F F SET)"
Nmap["BOOL.to..exists.colors.PRED.LAMBDA."] <- "BOOL to (exists-colors PRED-LAMBDA)"
Nmap["BOOL.to..exists.shapes.PRED.LAMBDA."] <- "BOOL to (exists-shapes PRED-LAMBDA)"
Nmap["BOOL.to..exists.sizes.PRED.LAMBDA."] <- "BOOL to (exists-sizes PRED-LAMBDA)"
Nmap["OBJ..BOOL.to.BV"] <- "F to BV"

Nmap["BOOL.to..OBJxOBJ..BOOL.SET.OBJ.SET.OBJ."] <- "BOOL to (RELN OBJECT OBJECT)"

Nmap["SET.to.S"] <- "SET to S"

D$name <- Nmap[D$name]

             
D <- D[order(D$name),]

# D$name <- ifelse(D$name == "FUNCTION.to..lambda..S...lambda..x...falsethunk...", "FUNCTION to (lambda S x true)", as.character(D$name))
# D$name <- ifelse(D$name == "FUNCTION.to..lambda..S...lambda..x...truethunk...", "FUNCTION to (lambda S x false)", as.character(D$name))
# D$name <- ifelse(D$name == "FUNCTION.to..lambda..S...lambda..x..BOOL..", "FUNCTION to (lambda S x BOOL)", as.character(D$name))
# D$name <- gsub("\\.to\\.", " to ", D$name)
# D$name <- gsub("\\.to\\.", " to ", D$name)

# D$name <- gsub("OBJ\\.\\.BOOL", "F", D$name)
# D$name <- gsub("SET\\.OBJ", "OBJECT", D$name)
# D$name <- gsub("\\.\\.$", "", D$name) # trailing ..
# D$name <- gsub("\\.\\.", " ", D$name) #
# D$name <- gsub(" \\.", " (", D$name) #
# D$name <- gsub("\\.$", ")", D$name) #
# D$name <- gsub("\\.", " ", D$name) #
#  
# D <- D[ order(as.character(D$name)),]
## Plot it
postscript(paste("grammar-",language,".eps",sep=""), width=7, height=10)
par(mar=c(5,20,1,1)) #par(mar=c(5,18,3,3))
b <- barplot( D$mean, horiz=T, names=D$name, las=2, border=0, col="white", xlim=c(0,5), xlab="Dirichlet parameter", main="")#main=toupper(language))
bar.y <- b 
arrows(D$hpd.lower, bar.y, D$hpd.upper, bar.y, angle=90, length=0.08, code=3)
# errbar(D$name, D$mean, D$hpd.lower, D$hpd.upper, add=TRUE, cex=1.5,lwd=2) # A funny format for horizontal bars, taking the grouping variable
points(D$mean, bar.y, col=1, cex=1.5, pch=20)
points(D$MAP, bar.y, col=2, cex=1.5, pch=20)
abline(v=0:5, lty=2)
dev.off()




