
# This makes a big barplot for all of the SFL concepts 
# A new version of crame plot. See Aug 6 version in Old for th eold vertical one.

##TODO
## Jan 21 Maybe sort by overall performance?


library(Hmisc) 
library(graphics) # for arrows
library(plotrix) # for pie chart

source("HSFL-Shared.R")

# Analyze Turk demographic data

d <- load.subject.data(print.summary=TRUE, remove.duplicate.concepts=FALSE) # don't remove duplicates because we need them to line everything up

# draw an arrow for each of these
percents <- c(0.25, 0.5, 0.75, 1.0) # This MUST start with something greater than 0.0 and end with 1.0
total.items <- 25

# create a matrix, one column for each concept
# and oen row for each percent. Then we arrow-plot that matrix
acc.matrix <- NULL
lower.matrix <- NULL # the lower 95% confidence
upper.matrix <- NULL # the upper 95% confidence

for(pct.ind in 1:length(percents)) {
	pct <- percents[pct.ind]
	
	minpct <- 0
	if(pct.ind > 1) { minpct <- percents[pct.ind-1]}

	sub <- d[(d$set.number <= pct*total.items) 
		  & (d$set.number > minpct*total.items),]

	# the mean value
	ag <- aggregate.binomial(sub$is.correct, by=list(sub$concept), names=c("concept"))

	if(nrow(ag) == 0) { 
		ag$p <- NA
		ag$lower <- NA
		ag$upper <- NA
	}

	acc.matrix <- rbind(acc.matrix, ag$p)
	lower.matrix <- rbind(lower.matrix, ag$lower)
	upper.matrix <- rbind(upper.matrix, ag$upper)

	if(is.null( colnames(acc.matrix))) {
		colnames(acc.matrix) <- ag$concept
	}
	else { 
		if( ! all( colnames(acc.matrix) == ag$concept)) { # check to be surewe are sitll adding rows in the right order
			print("PROBLEM LINING UP COLUMN NAMES"); break;
		}	
	}

}

acc.matrix <- rbind( 0, acc.matrix) # put a 0 on the start
lower.matrix <- rbind( NA, lower.matrix) # put a 0 on the start
upper.matrix <- rbind( NA, upper.matrix) # put a 0 on the start

#	the chance if you guessed at chance by sampling from the right distribution
ag.chance <- aggregate( d$right.answer, by=list(d$concept), function(x){mean(x,na.rm=T)})
names(ag.chance) <- c("concept", "chance.pct")
chance <-ag.chance$chance.pct*ag.chance$chance.pct + (1-ag.chance$chance.pct)*(1-ag.chance$chance.pct); names(chance) <- ag.chance$concept


# Which ones are boolean and nonboolean
boolean.concepts <- names(concept.class)[ (! is.na(concept.class)) & concept.class=="B"]
nonboolean.concepts <- names(concept.class)[ (! is.na(concept.class)) & (! concept.class=="B")]

# Now sort by last bin performance
sorted.concepts <- colnames(acc.matrix)[order(colSums(acc.matrix))]
sorted.concepts <- sorted.concepts[ !is.na(concept.class[sorted.concepts])] # remove duplicates

break1 <- floor(length(sorted.concepts) * 0.333)
break2 <- floor(length(sorted.concepts) * 0.666)

S1 <- sorted.concepts[ 1:break1]
S2 <- sorted.concepts[ (1+break1):break2]
S3 <- sorted.concepts[ (1+break2):length(sorted.concepts) ]


## NOTE: For now this MUST be consecutive (since we compute y based on it)
Si <- 3
for(cons in list(S1, S2, S3)) { #list(1:10, 11:20, 21:30, 31:40)) {

#	x11()
 	postscript(paste("horiz-crane-", Si, ".eps", sep=""), height=10, width=12)
# 	bitmap(paste("horiz-crane-", Si, ".png", sep=""), height=7, width=10, taa=2, gaa=4, res=200)
	par(mar=c(6,30,2,1), xpd=T, font=T)
	# to make the names have concept number, names=paste(toupper(concept.desc[cons]),cons)
	y <- barplot(rep(0, length(cons)), xlim=c(0.5, 1.01), names=toupper(concept.desc[cons]), las=1, cex.names=0.75, cex.lab=1.0, border=0, col=0, font=2, horiz=TRUE, xlab="Proportion Correct", cex.axis=1.3, cex.lab=1.5)

	rows <- c(2,3,4,5) # plot these rows of acc.matrix

	yi <- 1 # keep track of where we are in y, in case the rows aren't consecutive
	for(c in cons) {

		linecol <- "#000000"
		if(is.element(c, boolean.concepts)) { linecol <- "#009900"}

		# draw from min to max
		segments( min(acc.matrix[rows,c]), y[yi], max(acc.matrix[rows,c]), y[yi], lwd=3, col=linecol)
		points( min(acc.matrix[rows,c]), y[yi], pch=1, cex=1.5)
		points( max(acc.matrix[rows,c]), y[yi], pch=19, cex=1.5)
		arrows( (max(acc.matrix[rows,c]) + min(acc.matrix[rows,c]))/2-0.02, y[yi], (max(acc.matrix[rows,c]) + min(acc.matrix[rows,c]))/2 + 0.01, y[yi], lwd=2, col=linecol, length=0.1)

		# now add each pie chart. Cute
# 		for(ri in 1:length(rows)) {
# 			floating.pie(acc.matrix[rows[ri],c], y[yi], c(1-percents[rows[ri]-1]*0.9999999, percents[rows[ri]-1]*0.999999), radius=0.007, col=c("#ffffff", "#000000"),edges=20)		
# 		}

		yi <- yi + 1;
	}
	
	# and plot chance
# 	points(chance[cons], y, col=4, pch=20, cex=2)
	segments(chance[cons], y-0.5, chance[cons], y+0.5, col=4, pch=20, cex=2, lwd=2)
	
	segments(1,0,1,max(y)) # add this line to make comparison easier

# 	legend(x=-1, y=0, c("First 25%", "Last 25%", "Chance"), pch=c(20,1,20), col=c(1,1,4), cex=1.5)

	dev.off()

	Si <- Si - 1

}





##########################################
####### Now we'll analyze some rbcl comparisons

compare.concepts <- function(a,b) {
	q <- d[ d$concept == a | d$concept == b,]
	
# 	q.ag.chance <- 

	f <- glm( is.correct ~ set.number*concept, data=q, family="binomial")
	cf <- coef(summary(f))
	cf#cf[nrow(cf),] # give me back the interaction term
}



