;; Shared functions and data for all parts of the SFL
(import (rnrs))

;; this must be set to tell it where the Thesis libraries live. 
;(define THESIS-SCHEME-PATH "/home/piantado/Desktop/mit/Libraries/r6rs-scheme/")

(define C-External-Path "./C/parameter-mcmc.so.0")
(load "HSFL-Languages.scm")
(load "HSFL-Concepts.scm")

(define debug-flag 0) ;; you can set this and conditionally display things

(define/log! MAX-DEPTH 8); The deepest our trees can get, eevrything deeper has 0 probability. We must generate to this+2 to avoid hanging on insert/delete

;; For quantifiers
(define/log! A-PREDICATE man?*)
(define/log! B-PREDICATE chef?*)

(define/log! ALPHA 0.75)
(define/log! BASERATE 0.5)
(define/log! PRIOR-TEMPERATURE 1.0)

(define/log! PCFG-TEMPERATURE 1.0);
(define/log! LL-BETA-POSITIVE -0.2);
(define/log! LL-BETA-NEGATIVE -0.2);



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Define our concepts and some classes of concepts
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; This is *literally* all concepts. But we ran some twice so use the version below
;(define ALL-CONCEPTS '("hg01" "hg02" "hg03" "hg04" "hg05" "hg06" "hg07" "hg08" "hg09" "hg10" "hg11" "hg12" "hg13" "hg14" "hg15" "hg16" "hg17" "hg18" "hg19" "hg20" "hg21" "hg22" "hg23" "hg24" "hg25" "hg26" "hg27" "hg28" "hg29" "hg30" "hg31" "hg32" "hg33" "hg34" "hg35" "hg36" "hg37" "hg38" "hg39" "hg40" "hg41" "hg42" "hg43" "hg44" "hg45" "hg46" "hg47" "hg48" "hg49" "hg50" "hg51" "hg52" "hg53" "hg54" "hg55" "hg56" "hg57" "hg58" "hg59" "hg60" "hg61" "hg62" "hg63" "hg64" "hg65" "hg66" "hg67" "hg68" "hg69" "hg70" "hg71" "hg72" "hg73" "hg74" "hg75" "hg76" "hg77" "hg78" "hg79" "hg80" "hg81" "hg82" "hg83" "hg84" "hg85" "hg86" "hg87" "hg88" "hg89" "hg90" "hg91" "hg92" "hg93" "hg94" "hg95" "hg96" "hg97" "hg98" "hg99" "hg100" "hg101" "hg102" "hg103" "hg104" "hg105" "hg106" "hg107" "hg108" "hg109" "hg110"  "hg111"  "hg112" 
;                               )) 

(define ALL-CONCEPTS '("hg01" "hg02" "hg03" "hg04" "hg05" "hg06" "hg07" "hg08" "hg09" "hg10" "hg11" "hg12" "hg13" "hg14" "hg15" "hg16" "hg17" "hg18" "hg19" "hg20" "hg21" "hg22" "hg23" "hg24" "hg25" "hg26" "hg27" "hg28" "hg29" "hg30" "hg31" "hg32" "hg33" "hg34" "hg35" "hg36" "hg37" "hg38" "hg39" "hg40" "hg41" "hg42" "hg43" "hg44" "hg45" "hg46" "hg47" "hg48" "hg49" "hg50" "hg51" "hg52" "hg53" "hg54" "hg55" "hg56" "hg57" "hg58" "hg59" "hg60" "hg61" "hg62" "hg63" "hg64" "hg65" "hg66" "hg67" "hg68" "hg71" "hg72" "hg73" "hg74" "hg75" "hg76" "hg77" "hg78" "hg79" "hg80" "hg81" "hg82" "hg83" "hg84" "hg85" "hg86" "hg87" "hg88" "hg89" "hg90" "hg91" "hg93" "hg94" "hg95" "hg96" "hg97" "hg98" "hg99" "hg100" "hg101" "hg102" "hg103" "hg104" "hg105" "hg106" "hg108" "hg109" "hg110"  "hg111"  "hg112" 
                               )) 


(define BOOLEAN-CONCEPTS '( "hg01" "hg02" "hg03" "hg04" "hg05" "hg06" "hg07" "hg08" "hg09" "hg10" "hg11" "hg12" "hg13" "hg14" "hg15" "hg16" "hg18" "hg19" "hg20" "hg21" "hg22" 
                               "hg23" "hg24" "hg25" "hg77" "hg78" "hg79" "hg80" "hg81" "hg82" "hg83" "hg84" "hg85" "hg86"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Functions and definitions for objects
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; set up the object space
(define possible-SFL-OBJECT-shapes '(circle triangle rectangle))
(define possible-SFL-OBJECT-colors '(blue green yellow))
(define possible-SFL-OBJECT-sizes  '(1 2 3))
(define possible-SFL-OBJECT-bgcolors '(3)) ;; For this, let's fix the bgcolor
(define possible-SFL-OBJECT-textures '(solid))

;; a function to sample an object uniformly
(define (sample-SFL-OBJECT)
  (make-SFL-OBJECT-nextid (sample-uniform possible-SFL-OBJECT-shapes)
                   (sample-uniform possible-SFL-OBJECT-colors)
                   (sample-uniform possible-SFL-OBJECT-sizes)
                   (sample-uniform possible-SFL-OBJECT-bgcolors)
                   (sample-uniform possible-SFL-OBJECT-textures)  ))

;; draw k samples of unique objects
(define (sample-unique-SFL-OBJECTs k) 
    (if (= k 0)
        '()
        (let ((s (sample-unique-SFL-OBJECTs (- k 1))))
          
          (let loop ((x (sample-SFL-OBJECT)))
            (if (or-list (map (lambda (y) (eqv-SFL-object* y x))
                            s)) ;; if x is the same as anything else
                (loop (sample-SFL-OBJECT))
                (cons x s))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; External C library
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define C-parameter-mcmc (find-lib (lambda (path) (dlopen path #t #f)) (list C-External-Path) pointer?))

;; --------------------------------------------------------------------------------------------
;; Using this external library is about 10X faster when using the N^2 algorithm, with lldecay, 
;; and about 2X *slower* when there is no ll decay
(define running_ll_Cexternal
  ((make-c-callout 'pointer '(double double double double unsigned-int pointer pointer pointer)) 
   (dlsym C-parameter-mcmc "running_ll_Cexternal")))
(define running-ll-C ;; a wrapper that takes cblocks and manages memory
  (lambda (a b llbp llbn len resp_cb cor_cb which_cb)
    (let* ((ptr (running_ll_Cexternal a b llbp llbn len resp_cb cor_cb which_cb))
           (out (cblock->list ptr 'double len)))
     ; (edisplayn out)
      (c_free ptr)
      out)))


;; --------------------------------------------------------------------------------------------
;; Free memory
(define c_free
  ((make-c-callout 'void '(pointer)) (dlsym C-parameter-mcmc "c_free")))

;; a helper
(define (to01 x) (if (eqv? x #t) 1 0))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Functions for managing the concept data
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;; this defines cl2data, which is the data for each concept
;; NOTE: This uses CONCEPTS and LISTS which are assumed to have been
;; defined somewhere
(define/eval define-concept-data
  '(
   ;; load up all the data for each concept,list
   (define cl2data (make-equal-hashtable))
   
   (for c in CONCEPTS
     (for l in LISTS
       (define input-file (open-input-file (get-concept-filename c l)))
       (define in-header (read-line input-file))
       (define all-data (first-n (process-file input-file read-concepts-line) ;; read everything else
                                 (+ 1 (apply max DATA-AMOUNTS))))
       (close-input-port input-file)
       
       (hashtable-set! cl2data (list c l) all-data)
       ))
   
   ))


;; returns a single data point list for a line of the file
(define (read-concepts-line x)
  (let* ((parts (str-split x #\tab))
         (correct (string->S-expression (car parts))))
    (cons correct (list (map get-SFL-OBJECT-from-string (cdr parts))))))

;; return the file name
(define (get-concept-filename c l) 
  (string-append CONCEPT-DIRECTORY "CONCEPT_" c "__LIST_" l ".txt"))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Functions for managing human data
;;;; These are mostly macros that will insert the right definitions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; For this, we want a hash from (concept list) to item-respn stacked, with
;; which_item, human_true, human_false, correct
;; which are all C vectors

(define input-human-responses (open-input-file IN-HUMAN-RESPONSES))

;; these map (concept list) to a list of when humans are true
(define cl2human-true (make-equal-hashtable))
(define cl2human-false (make-equal-hashtable))
(define cl2correct (make-equal-hashtable))
(define cl2which-item (make-equal-hashtable))
(define cl2data-N (make-equal-hashtable))
(define cl2running-ll (make-equal-hashtable))

(process-file 
 input-human-responses 
 (lambda (l)
   (let* ((parts (str-split l #\tab))
          (numparts (map string->number parts)) ;; concept is not a number so access it via parts, not numparts
          (key (list (nth 1 parts) ;; the key is (concept listnumber)
                     (nth 2 parts))))
     
     ;; add one more since we predict on one more than max DATA-AMOUNTs
     ;; it is important that we do this since thats how use this to count the number of data points (for average ll)
     (hashtable-append! cl2which-item key (list (nth 3 numparts)))
     (hashtable-append! cl2human-true key (list (nth 5 numparts)))
     (hashtable-append! cl2human-false key (list (nth 6 numparts)))
     (hashtable-append! cl2correct key (list (nth 7 numparts)))))) ;;))

;; compute the overall total number of responses
(define TOTAL-NUMBER-OF-RESPONSES (apply + 
                                         (formap k in (vector->list (hashtable-keys cl2human-true))
                                                 (+ (apply + (hashtable-ref cl2human-true k null))
                                                    (apply + (hashtable-ref cl2human-false k null))))))

;; and now set cl2data-N, the number of data points. 
;; we can pul lthis from any of the above, since they should all be the same!
(for k in (vector->list (hashtable-keys cl2human-true))
  (hashtable-set! cl2data-N k (+ (apply + (hashtable-ref cl2human-true k null))
                                 (apply + (hashtable-ref cl2human-false k null)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Functions for managing lists of hypotheses
;;;; This stores them in the format of delabel<tab>PCFG-NODE
;;;; and reads back only the PCFG-NODE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                  

(define (get-hypotheses-file prefix lang c l type) 
  (string-append prefix "hypotheses__LANGUAGE_" lang "__CONCEPT_" c "__LIST_" l "__TYPE_" type ".txt"))

;; returns a list of hypotheses (trees only) for c l and type
(define (load-hypothesis-list prefix lang c l type mytop)
  (let ((tmphash (make-PCFG-hashtable)))
    (hashtable-loadfile! tmphash (get-hypotheses-file prefix lang c l type))
    (hashtable-filter! tmphash (lambda (k v) (< v mytop))) ;; throw  out low prob 
    (vector->list (hashtable-keys tmphash))))
     
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Functions for getting responses from a hypothesis
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                  

;; Each data point will be a list of the form ( (correct-answer-1 correct-answer-2) (obj-1 obj-2 obj-3 obj-4))
;; ie, a list for objects and a list for correct answers

;; gets the function response for an expression, not a PCFG-NODE tree
(define (get-function-response-expr expr sets)
  
  ;; F will store the function, which always, sometimes vacuously, takes a context set S
  ;; to simply its evaluation. This means we have to look to see if it already does, and if not, 
  ;; add another lambda
  (if (= QUANTIFIER-FORMAT 1)
      (let ((F (eval expr (interaction-environment)))) ;; if we do things as with quantifiers, (lambda (S A B) ...)
        (map (lambda (S) (list (F S (filter A-PREDICATE S) (filter B-PREDICATE S)))) ;; return a list since that's what everything handles
             sets))      
      (let ((F (cond ((not (list? expr)) ;; Set function format
                      (eval `(lambda (S) ,expr) (interaction-environment)))
                     ((and (list? (third expr)) (eqv? (car (third expr)) 'lambda)) ;; Else it better be a lambda. If it has two lambdas
                      (eval expr (interaction-environment)))
                     (#t
                      (eval `(lambda (S) ,expr) (interaction-environment))))))
        (map (lambda (S)
               (let ((Fs (F S)))
                 (map Fs S)))
             sets))))

;; This returns a list of the function represented by H applied to every set in sets
;; Note that sets MUST be a list of sets
;; This returns a list of responses (which are lists, one for each item), one for each set
(define (get-function-response H sets)
  (get-function-response-expr (PCFG-NODE-delabel H) sets))

(define (get-single-function-response-expr expr set)
  (car (get-function-response-expr expr (list set))))

;; a wrapper for dealing with the case where you want to eval on only one set
(define (get-single-function-response H set)
  (car (get-function-response H (list set))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Computation of running-ll 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;           


;; this computes a list of running ll. That is your score on all previous
;; sets, but weighting them by the power law decay llBeta
(define (running-ll-SCHEME PARAM-alpha PARAM-baserate PARAM-llBeta response correct which-item)
  
; (disp "Correct:")(disp correct) (newline)
;  (disp "Response:")(disp response) (newline) 
;  (displayn PARAM-baserate)
;  (displayn PARAM-alpha)  
  
  (decay-running-ll (map (lambda (c r) (log (+ (if (eqv? c r) PARAM-alpha 0.0) ;; if right
                                               (* (- 1.0 PARAM-alpha) ;; sample based on chance
                                                  (if r PARAM-baserate (- 1.0 PARAM-baserate))))))
                         correct
                         response)
                    which-item
                    PARAM-llBeta))

(define (decay-running-ll pointwise-ll which-item llBeta) 
  
  (let ((len (length pointwise-ll))
        (pw (apply vector pointwise-ll))
        (wi (apply vector which-item))
        (out (make-vector (length which-item) 0))
        
        
        ;; find the first thing that's not the same item as the first item
        (start (list-index (lambda (k) (not (= k (car which-item))))
                           which-item)))
   ;; C-ish code, for jerks. 
  (let loop-i ((i start))
    (let loop-j ((j 0))
      
      (vector-set! out i (+ (vector-ref out i)
                            (* (vector-ref pw j)
                               (expt (+ 1 (- (vector-ref wi i)
                                             (vector-ref wi j)))
                                     llBeta))))      
      
      (if (and (< (+ j 1) i)
               (not (= (vector-ref wi (+ j 1)) ;;if we are not on your item, then loop 
                       (vector-ref wi i))))
          (loop-j (+ j 1))))
    (if (< (+ i 1) len) (loop-i (+ i 1))))
  
  (vector->list out)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Functions for computing priors
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                  

;; assumes that p is an unnormalized, non-log distribution in a list, and cnt is the counts 
;; of how often each has occurred
(define (score-p-cnt p cnt)
  (let ((n (log (apply + p))))
    (apply + (map (lambda (ci pi) (if (> ci 0)
                                      (* ci (- (log pi) n))
                                      0.0))
                  cnt
                  p))))

;; log(  prod_i Gamma(c_i) / Gamma(sum(c_i))  )
;(define (lg-RR-beta c)
;  (- (apply + (map gsl-lngamma c)) 
;     (gsl-lngamma (sum c))))

;; assumes that p is an unnormalized, non-log distribution in a list, and cnt is the counts 
;; of how often each has occurred. This scores according to a RR prior, with
;; cnt a dirichlet
;(define (score-p-RR p cnt)
;  (- (lg-RR-beta (map + cnt p)) ;; TODO: DO WE NeED A NORMALIZER HERE?
;     (lg-RR-beta p))) 

;(define (score-p-RR-flat p cnt) ;; this has flat priors
;  (- (lg-RR-beta (map (lambda (x) (+ x 1)) cnt)) ;; TODO: DO WE NeED A NORMALIZER HERE?
;     (lg-RR-beta (map (lambda (x) 1) cnt))))    

;; This takes a scorer, which maps a list of params and counts
;; to a probability. It also takes hypothesis rule count and a grammar parameter
(define (PCFG-prior-score-count scorer hrc gp)
  (apply + (map (lambda (glst hlst) (+ (if (> (first (first hlst)) 0.0) ;; if we drew a bound variable
                                           (+ (* (first (first hlst)) ;; how many times we drew a bound variable
                                                 (- (log (first glst)) (log (apply + glst)))) ;; the normalized prob of drawing the bv 
                                              (second (first hlst))) ;; prob of all bound variables
                                           0)                                      
                                       (scorer (cdr glst) (cdr hlst))))
                gp
                hrc)))


(define (rule-count-PCFG-prior hrc gp)
  (PCFG-prior-score-count score-p-cnt hrc gp))

(define (rule-count-RR-prior hrc gp)
  (PCFG-prior-score-count RR-parameterized-scorer hrc gp))

(define (rule-count-RR-prior-flat hrc gp) ;; ignnore gp
  (PCFG-prior-score-count RR-flat-scorer hrc gp))


;; I think these should return *functions* that compute the priors, integrating the parameters, so that we don't ahve to pass them
;; around. Is that actually a good idea?

;; check if a tree is valid according ot our constraints
;; For now, just not too deep
(define (is-valid-tree H)
  (and (<= (PCFG-NODE-depth H) MAX-DEPTH)))

(define (compute-prior-UNIFORM H)
  (if (is-valid-tree H)
      0.0
      -inf.0))

;; a single parameter, the temperature
(define (compute-prior-FLAT-PCFG H T)
  (if (is-valid-tree H)
      (/ (PCFG-log-probability H) T)
      -inf.0))

;; this get a vector corresponding to all of the prior features of h
;; right now, this is each of the primitives used
;; and then whether it is setwise-constant (exlcuidng #t and #f)
(define (get-prior-feature-vector hypothesis)
  
  (let* ((my-count (make-equal-hashtable)))
    (for-each (lambda (p) ;; load this into the hashtable
                (hashtable++ my-count p))
              (PCFG-get-functions-used hypothesis))
    
    (map (lambda (r) (hashtable-ref my-count r 0))
         FUNCTIONS-TO-TRACK)))

;; Compute the prior for a hypothesis. For now, just the PCFG probability
(define (compute-PCFG-prior H)
  (if (> (PCFG-NODE-depth H) MAX-DEPTH)
      -inf.0  
      (* (/ 1.0 PCFG-TEMPERATURE) 
         (PCFG-log-probability H))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Functions for the likelihood
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                  


;; Here D is  alist of data points, which each data point is a list of ( <correct-response-list> <object-list>)
;; Importantly, D is ordered with the most recent appearing *last*

;; we cache the computation of the likelihood
;(define likelihood-cache (make-equal-hashtable 10000))
;(define (clear-likelihood-cache)
;  (hashtable-clear! likelihood-cache))
;
;;; This is an unsafe version which DOES NOT hash the data, meaning that
;;; it will incorrectively give you the wrong answer when you change the data
;;; this means when you change data, you shoudl call (clear-likelihood-cache)
;(define (compute-likelihood-cached-UNSAFE D H) 
;  (define expr (PCFG-NODE-delabel H))
;  (define ret (hashtable-ref likelihood-cache expr ':NOKEY))
;  (if (eqv? ret ':NOKEY)
;      (let ((ll (compute-likelihood D H))) ;; compute it and store it
;        (hashtable-set! likelihood-cache expr ll)
;        ll)
;      ret))


;;; This takes a hypothesis
(define (compute-likelihood alpha baserate llbeta-p llbeta-n  D H)
  (compute-likelihood-listy alpha baserate llbeta-p llbeta-n
                            D
                            (get-function-response H (map second D))))

;;; This takes a hypothesis
(define (compute-likelihood-default D H)
  (compute-likelihood-listy ALPHA BASERATE LL-BETA-POSITIVE LL-BETA-NEGATIVE
                            D
                            (get-function-response H (map second D))))

;; This takes a list of responses. 
;; Here D is a list of all the correct answers and sets, and Hlist is the function's responses
;; and Hlist is a list of that hypothesis's responses
(define (compute-likelihood-listy alpha baserate llbeta-p llbeta-n D Hresp) 
  (let ((N (length Hresp))
        (corrects (map first D))
        (ll 0))    
    (for-each (lambda (cor resp)
                (for-each (lambda (c r)
                            (set! ll (+ ll (*  (expt (+ N 1) (if c llbeta-p llbeta-n)) ;; the likelihood decay
                                               (log (+ (if (eqv? c r) alpha 0)
                                                       (* (- 1 alpha) (if r baserate (- 1 baserate)))))))))
                          cor
                          resp)
                (set! N (- N 1))
                )
              corrects
              Hresp)
    ll))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Stuff for finding vacuous and bizzare hypotheses
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                  


;; This adds the top in top-mcmc to the hash hsh 
;; This also computes the posterior on everything in tn-initializer
;; and uses that to initialize top n (thus you must do better than them)
;(define (add-top-mcmc! hsh D steps temperature topcount tn-initializer)
;  
;  (define tn (construct-TOP-N-UNIQUE topcount make-PCFG-hashtable)) ; the top hypotheses
;  (define initial-H (PCFG-generate SFL-grammar 'FUNCTION MAX-DEPTH))
;  
;  ;; add the initializer
;  (for tni in tn-initializer
;       (potentially-insert-TOP-N! tn
;                                  tni
;                                  (+ (compute-PCFG-and-feature-prior tni)
;                                     (compute-likelihood D tni))))
;  ;; and do some MCMC
;  (metropolis-hastings-simple! (make-MH-STATE initial-H 
;                                              (compute-PCFG-and-feature-prior initial-H)
;                                              (compute-likelihood D initial-H))
;                               steps
;                               0
;                               temperature
;                               compute-PCFG-and-feature-prior ;; a function of hypothesis
;                               compute-likelihood ;; a function of data and hypothesis
;                               D
;                               (lambda (h)  (PCFG-NODE-resample SFL-grammar h (+ MAX-DEPTH 2))) 
;                               (lambda (mhs) ;; call this on every sample
;                                 (potentially-insert-TOP-N! tn 
;                                                            (MH-STATE-hypothesis mhs) 
;                                                            (MH-STATE-posterior mhs)))
;                               PCFG-NODE-eqv?
;                               null
;                               )
;  
;  ;; add the top ones we found
;  (map (lambda (h) (hashtable-set! hsh h 1))
;       (filter PCFG-NODE? (vector->list (TOP-N-UNIQUE-objects tn)))))


;; does this hypothesis always give the same answer? 
;; This automatically checks on all-data
;; NOTE: This is specific to T, F, undef responses
;; NOTE: This has the same type interface as sample-predictive, and
;; should just be called in there after checking REMOVE-VACUOUS-HYPOTHESES

;; Are we true-thunk, false-thunk, or any of their incarnations?
(define (is-binary-thunk s)
   (let ((dl (PCFG-delabel s)))
     (or (equal? dl '(lambda (x) (truethunk)))
            (equal? dl '(lambda (x) (falsethunk)))
            (equal? dl 'truethunk)
            (equal? dl 'falsethunk)
            (equal? dl '(lambda (x) #t))
            (equal? dl '(lambda (x) #f)))))

(define vacuous-hash (make-PCFG-hashtable)) ;; this caches this function so that we don't have to re-call it for every alpha and beta

(define (is-hypothesis-vacuous? s data)
  (if (is-binary-thunk s)
      (= VACUOUS-AND-CONSTANT-INCLUDE-BINARY-THUNKS 1) ; if we are a binary thunk, then we are counted as vacuous iff this flag is 1
      (call/cc (lambda (exit) (is-hypothesis-vacuous?-cc s exit data)))))
(define (is-hypothesis-vacuous?-cc s exit data)
  
  (define Tcnt 0)
  (define Fcnt 0)
  (define Ucnt 0)
  
  (map (lambda (d)
         (map (lambda (r)
                (cond ((eqv? r '#t) (if (> (+ Fcnt Ucnt) 0) 
                                        (exit #f)
                                        (set! Tcnt 1)))
                      ((eqv? r '#f) (if (> (+ Tcnt Ucnt) 0) 
                                        (exit #f)
                                        (set! Fcnt 1)))
                      ((eqv? r U) (if (> (+ Fcnt Tcnt) 0) 
                                      (exit #f)
                                      (set! Ucnt 1)))
                      ))
              (get-single-function-response s (second d)) )
         )
       data)
  
  ;; else we are vacuous
  (exit #t))

;; Check if this function is always the same on every element of a set
;; e.g. Everything iff there is a triangle
(define (is-hypothesis-setwise-constant? s data)
  (if (is-binary-thunk s)
      (= VACUOUS-AND-CONSTANT-INCLUDE-BINARY-THUNKS 1)
      (call/cc (lambda (exit) (is-hypothesis-setwise-constant?-cc s exit data)))))
(define (is-hypothesis-setwise-constant?-cc s exit data)  
  (map (lambda (d)
         (let ((mysum (apply + (map (lambda (r) (if r 1 0))
                            (get-single-function-response s (second d)) ))))
           ;(displayn (list mysum (length (second d)) (PCFG-delabel s)))
           (if (and (not (= mysum (length (second d)))) ;; if we don't see all T or all F
                    (not (= mysum 0)))
               (exit #f))))
       data)
  
  ;; else we are setwise constant 
  (exit #t))


; This function takes a list of PCFG-NODEs and some data and uses the data to collapse
; hypothese that give the same result on the data. These hypotheses are collapsed to their
;; most concise form. 
(define (collapse-crazy-hypotheses-to-concise hypotheses data)
  
  ;; this will hash a list of responses like '(#t #f #f #f #t ..) to a hypothesis
  ;; which represents the most concise form of those hypotheses yet. This hash function
  ;; can just do something simple like add up all the #ts
  (define response2besthyp (make-hashtable (lambda (l) (sum (map (lambda (x) (if (eqv? x #t) 1 0))
                                                              l)))
                                        equal? ))
  
  (for-each (lambda (h)
              (let* ((responses (append-map (lambda (di)
                                             (get-single-function-response h (second di)))
                                           data))
                     (curbest (hashtable-ref response2besthyp responses 'NONE)))
                (if (or (eqv? curbest 'NONE) ;; if this pattern hasn't been seen or we have higher prior
                        (> (PCFG-log-probability h)
                           (PCFG-log-probability curbest)) )
                    (hashtable-set! response2besthyp responses h) ;; save!
                    )))
            hypotheses)
  
  ;; and return 
  (hashtable-values response2besthyp))
  


;; Is this perfect on the data?
(define (is-hypothesis-perfect? s data)
  (call/cc (lambda (exit) (is-hypothesis-perfect?-cc s exit data))))
(define (is-hypothesis-perfect?-cc s exit data)
    
  (map (lambda (d)
         (map (lambda (c r) (if (not (eqv? c r)) (exit #f)))
              (first d)
              (get-single-function-response s (second d)) ))
       data)
  ;; else we are perfect
  (exit #t))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Misc
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;                  

