(import (rnrs files (6)))

;; This implements a simpler learning model that doesn't do data analysis, but only does inference over expressions.  
(load "/home/piantado/Desktop/mit/Libraries/LOT-lib/LOT-lib.scm")
(load "HSFL-General.scm") ;; shared parts of SFL

(define/log! IN-TOP 250)
(define/log! IN-HYPOTHESES "mcmc-hypotheses/") ;; where the initializing hypotheses live
(define/log! IN-HYPOTHESIS-TYPE "RAW") ; "RAW" or "NONCRAZY"

(define/log! CONCEPTS BOOLEAN-CONCEPTS)
(define/log! LISTS (list "L1")); "L2"))

(define/log! LANGUAGE "Language-FOL-Other") ;; Language-Base, Language-Boolean, Language-No-Quantifiers
(define/log! MODEL "./models/parameterized-RR.scm")

(define/log! PARAM-alpha 0.5)
(define/log! PARAM-gamma 0.5)
(define/log! PARAM-llbeta -0.10)
(define/log! PARAM-prior-temperature 1.0)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Set up input and log it
(parse-command-line)

;; Fix IN-TOP to be a number
(if (string? IN-TOP)
    (set! IN-TOP (string->number IN-TOP)))

(define SFL-grammar (eval (string->symbol LANGUAGE) (interaction-environment)))

;; Randomize the RNG
(random-source-randomize! default-random-source) 

(define-concept-data)

(load MODEL)

(set-model) ;; This sets all the model parameters and make-model to the correct things for what was loaded just above

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Run MCMC to refresh model c, l according to our grammar
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;; Main MH loop
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;; Load the target concept
(define target (make-equal-hashtable))
(define approx (make-equal-hashtable))
(define (m-target c l) (hashtable-ref target (list c l) null)) ;; a separate model for each c and l
(define (m-approx c l) (hashtable-ref approx (list c l) null))

(for c in CONCEPTS
  (for l in LISTS
    
    (edisplayn "Loading " c "\t" l)
    
    (hashtable-set! target (list c l) (make-model))
    ((m-target c l) 'model-set-correct!   (apply append (map first (hashtable-ref cl2data (list c l) null))))
    ((m-target c l) 'model-set-which-item! (hashtable-ref cl2which-item (list c l) null))
    (for h in  (load-hypothesis-list IN-HYPOTHESES LANGUAGE c l IN-HYPOTHESIS-TYPE 100000)    ;; And load the hypotheses
      ((m-target c l) 'model-push-hypothesis! 
                      h (apply append (get-function-response h (map second (hashtable-ref cl2data (list c l) null))))))
    ((m-target c l) 'model-set-data! (hashtable-ref cl2data (list c l) null))     ;; and set the model data
    
    (hashtable-set! approx (list c l) (make-model))
    ( (m-approx c l) 'model-set-correct!   (apply append (map first (hashtable-ref cl2data (list c l) null))))
    ( (m-approx c l) 'model-set-which-item! (hashtable-ref cl2which-item (list c l) null))
    (for h in  (load-hypothesis-list IN-HYPOTHESES LANGUAGE c l IN-HYPOTHESIS-TYPE IN-TOP)    ;; And load the hypotheses
      ( (m-approx c l) 'model-push-hypothesis! 
                       h (apply append (get-function-response h (map second (hashtable-ref cl2data (list c l) null))))))
    ( (m-approx c l) 'model-set-data! (hashtable-ref cl2data (list c l) null))     ;; and set the model data
    
    
    ))

(define grammar-parameter (SFL-grammar 'get-flat-count-vector 1.0))

;; + + + + + + + + + + + + + + + + + + + + + + + + + + +  
;; Now actual mcmc
(for c in CONCEPTS
  (for l in LISTS
    
    (let ((target-prediction ( (m-target c l) 'model-prediction 
                               (vector PARAM-prior-temperature ; temperature ;; extract these from across parameters
                                       PARAM-alpha
                                       PARAM-gamma
                                       PARAM-llbeta)
                               grammar-parameter 
                               across-parameters))
          (approx-prediction ( (m-approx c l) 'model-prediction 
                               (vector PARAM-prior-temperature ; temperature ;; extract these from across parameters
                                       PARAM-alpha
                                       PARAM-gamma
                                       PARAM-llbeta)
                               grammar-parameter 
                               across-parameters)))
      ;(edisplayn target-prediction)
      ;(edisplayn approx-prediction)
      (displayn c 
                "\t" 
                l 
                "\t" 
                IN-TOP
                "\t"
                (sqrt (apply + (map (lambda (x y) (* (- x y) (- x y))) ;; RMSE
                                                  target-prediction
                                                  approx-prediction)))
                "\t"
                (sqrt (apply max (map (lambda (x y) (abs (- x y)))
                                      target-prediction
                                      approx-prediction)))
                "\t"
                (sqrt (mean (map (lambda (x y) (abs (- x y)))
                                 target-prediction
                                 approx-prediction)))
                "\t"
                (mean (map (lambda (x y) 
                             (if (>= x 1.0) (set! x (- 1.0 1e-6)))
                             (if (>= y 1.0) (set! y (- 1.0 1e-6)))
                             (if (<= x 0.0) (set! x 1e-6))
                             (if (<= y 0.0) (set! y 1e-6))
                             (+ (* x (- (log x) (log y)))
                                            (* (- 1.0 x) (- (log (- 1.0 x)) (log (- 1.0 y)))))) ;; mean KL responsewise
                           target-prediction
                           approx-prediction))
                ))
    
    
    ))
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;; END Main MH loop
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;(exit)
