(import (rnrs files (6)))

;; This implements a simpler learning model that doesn't do data analysis, but only does inference over expressions.  
(load "/home/piantado/Desktop/mit/Libraries/LOT-lib/LOT-lib.scm")

(define/log! OUT-PREFIX "run-local/SI/") ;; a directory which will be created, for our output files
(define LOG-OUT-FILE "/dev/null");(string-append OUT-PREFIX "/dev/null"))

(define/log! IN-TOP "100")
(define/log! IN-HYPOTHESES "mcmc-hypotheses/") ;; where the initializing hypotheses live
(define/log! IN-HYPOTHESIS-TYPE "RAW") ; "RAW" or "NONCRAZY"

(define/log! CONCEPT "hg63")
(define/log! LIST "L2")

(define/log! LANGUAGE "Language-LAM-CRD") ;; Language-Base, Language-Boolean, Language-No-Quantifiers

(define/log! PARAM-alpha 0.38) ; 0.75
(define/log! PARAM-gamma 0.5); 0.5
(define/log! PARAM-beta-positive -0.80) ;-.1
(define/log! PARAM-beta-negative -0.90);-.1
(define/log! PARAM-prior-temperature 4.3) ; 1.0

(define/log! DATA-AMOUNTS 24); IMMEDIATELY converted to a list below 0.. this

;; These are needed in HSFL-General, but have to be defined here so that we can parse them on the command line
(define/log! IN-HUMAN-RESPONSES "./Data-Scheme-HELD-OUT.txt") 
(define/log! CONCEPT-DIRECTORY "../concepts/") ;; where the data p<CONCEPT>.L<List>.txt lives
(define/log! QUANTIFIER-FORMAT 0) ; 1 = we are doing quantifiers, else set functions (set->subset)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Set up input and log it
(parse-command-line)

(set! DATA-AMOUNTS (seq 0 DATA-AMOUNTS 1)) 

(load "HSFL-General.scm") ;; shared parts of SFL

(define CONCEPTS (list CONCEPT)) ;; these are needed for HSFL-General
(define LISTS (list LIST))

;; Fix IN-TOP to be a string
(if (string? IN-TOP)
    (set! IN-TOP (string->number IN-TOP)))

(define SFL-grammar (eval (string->symbol LANGUAGE) (interaction-environment)))

;; Randomize the RNG
(random-source-randomize! default-random-source) 

(define-concept-data)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Here are variables that would be resumed
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(tabbycatn LOG-OUT-FILE "Initializing hypotheses.")
(define cl2allhyps (make-equal-hashtable)) ;;; a list of hypotheses (PCFG-NODEs) for each initializing concept
(define cl2initialhyps (make-equal-hashtable)) ;; save the initial hypotheses for when we refresh

(let ((c CONCEPT) (l LIST))
  (tabbycatn LOG-OUT-FILE "Reading data files " c " " l )
  (let ((hyplist (load-hypothesis-list IN-HYPOTHESES LANGUAGE c l IN-HYPOTHESIS-TYPE IN-TOP)))
    
    (hashtable-set! cl2allhyps (list c l) hyplist) ;; these are the ones 
    (hashtable-set! cl2initialhyps (list c l) hyplist) 
    
    ))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Run MCMC to refresh model c, l according to our grammar
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;; Main MH loop
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



(displayn "concept\tlist\tresponse.index\tposterior\tresponse\tpattern\thypothesis")

;; + + + + + + + + + + + + + + + + + + + + + + + + + + +  
(let ((c CONCEPT) (l LIST))
  
  (define hypotheses (hashtable-ref cl2allhyps (list c l) null))
  (define correct (map to01 (apply append (map first (hashtable-ref cl2data (list c l) null)))))
  (define correct_cb (list->cblock correct 'int))
  
  (define which-item_cb (list->cblock (hashtable-ref cl2which-item (list c l) null) 'int))
  (define len -1)
  ;; precompute each hypothesis' responses
  (define hyp2runningposterior (make-PCFG-hashtable))
  (define hyp2responses (make-PCFG-hashtable))
  (define hyp2profile (make-PCFG-hashtable)) ; we'll store a functional profile of this hypothesis
  (define grammar-parameters (SFL-grammar 'get-flat-count-vector 1.0))
  
  (for h in hypotheses
    (define responses (map to01 (apply append (get-function-response h (map second 
                                                                            (hashtable-ref cl2data (list c l) null))))))
    (define responses_cb (list->cblock responses 'int))
    
    (hashtable-set! hyp2profile h (string-hash (apply string-append (map number->string responses))))
    
    (hashtable-set! hyp2responses h responses)
    (hashtable-set! hyp2runningposterior 
                    h
                    (map (lambda (x) (+ (/ (SFL-grammar 'log-probability-scorer 
                                                        RR-parameterized-scorer 
                                                        (SFL-grammar 'get-rule-counts h) 
                                                        grammar-parameters)
                                           PARAM-prior-temperature) ;; prior
                                        x))
                         (running-ll-C PARAM-alpha ; alpha
                                       PARAM-gamma ; gamma
                                       PARAM-beta-positive; decay
                                       PARAM-beta-negative; decay
                                       (length responses)
                                       responses_cb
                                       correct_cb
                                       which-item_cb)))
    (set! len (length responses)) ;; how many data points
    (free responses_cb))
  
  ;; now loop over data, copmuting the likelihood
  (for d in (seq 0 (- len 1))
    (let* ((posteriors (map (lambda (h) (list-ref (hashtable-ref hyp2runningposterior h null)
                                                  d))
                            hypotheses))
           (responses (map (lambda (h) (list-ref (hashtable-ref hyp2responses h null)
                                                 d))
                           hypotheses))
           (Z (logsumexp-lazy posteriors)))
      (map (lambda (h p r)
             (displayn c "\t" l "\t" d "\t" (- p Z) "\t" r "\t" (hashtable-ref hyp2profile h 0) "\t\"" (PCFG-delabel h) "\""))
           hypotheses
           posteriors
           responses))))


;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;; END Main MH loop
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

(exit)
