

;; This is a version that just does logistic regression on accuracy


;; This takes care of all of the caching etc
;; and implements:

;; (model-push-hypothesis! h resp) -- pushes the hypothesis h with respones resp onto data
;; (model-set-correct! d) -- a set of data
;; (model-set-data! d)
;; (model-set-which-item! w) -- a list saying which item we are on
;; (model-prediction p) -- takes a parameter value and computes the model predicted proportions (possibly with fancy caching)
;; (get-lower-bound) -- retrieve bounds on p
;; (get-upper-bound) 
;; (get-initial-p) -- what p we can start with?

; http://community.schemewiki.org/?simple-object

; -----------------------------------------------------------------------------------------
;; Define across parameters (outside here, so that they are shared)

(define (set-model)
  (set! across-parameters (vector ))
  (set! across-proposers (vector ))
  
  (set! make-model make-model:logistic))
  
; -----------------------------------------------------------------------------------------

(define (make-model:logistic)
  
  (define META-description "Uniform over hypotheses." )
  (define META-notes       "(Uses C external library for running-ll)." )
  
  (define correct null) ;; stored in collapsed format
  (define which-item null)
  (define len 0)
  
  (define (initialize-model) 0)  
  (define (model-set-data! d) null)
   
  
  (define (get-initial-p)    (vector 0.0 0.0 0.0 ))
  (define (get-proposers) (vector (lambda (x) (normal-proposal x 0.1))
                                  (lambda (x) (normal-proposal x 0.1))
                                  (lambda (x) (normal-proposal x 0.1))))
  
  (define (model-set-correct! c) 
    (set! correct c)
    (set! len (length correct))
    )
  
  (define (model-set-which-item! w) 
    (set! which-item w)
    )
  
  ;; adds h to the model, with c and l
  (define (model-push-hypothesis! h resp)  null    )
  
  ;; --------------------------
  ;; Local helper functions
  
  (define (logistic x) (/ 1.0 (+ 1.0 (exp (- x)))))
  
  
  ;; --------------------------
  (define  (model-prediction p gp ap)
    
    (let ((PARAM-intercept   (vector-ref p 0))
          (PARAM-slope       (vector-ref p 1))
          (PARAM-istrue      (vector-ref p 2))) ;; the effect on accuracy of it being true/false
      
      
      ;;; Here our prediction is just the correct answer with 
      ;;; logit of slope+intercept
      (map (lambda (ci wi) (let ((a (+ PARAM-intercept 
                                       (* PARAM-slope wi) 
                                     (if (eqv? ci #t) PARAM-istrue 0.0))))
                           (if (eqv? ci #t) 
                               (logistic a)
                               (- 1.0 (logistic a)))))
         correct
         which-item
         )))
  
  (define model-entropy (lambda args 0))
  
   ;; ------------------------------------------------
  ;;; And done, so make this thing a class
  (lambda (message . args)
    (apply (case message 
             ((initialize-model) initialize-model) 
             ((get-initial-p) get-initial-p)
             ((get-lower-bound) get-lower-bound) 
             ((get-upper-bound) get-upper-bound) 
             ((get-proposers) get-proposers) 
             ((model-push-hypothesis!) model-push-hypothesis!) 
             ((model-set-correct!) model-set-correct!) 
             ((model-set-which-item!) model-set-which-item!) 
             ((model-set-data!) model-set-data!)
             ((model-prediction) model-prediction) 
             ((model-entropy) model-entropy)
             (else (die (string-append "Unknown Method! " (symbol->string message))))) 
           args))) 